(require [kodhy.macros [filt]])

(import
  re
  os
  time
  calendar
  csv
  sqlite3
  [unidecode [unidecode]]
  [cheat-importing.common [*]])

(setv csv-dir "/tmp/cct")
(setv zipstates-db-path "../data/zipstates.sqlite")
(setv target-db-path "/dev/shm/amcc.sqlite")

(setv commit-every 100,000)
(setv start-with None)

;; ------------------------------------------------------------
;; * Functions and macros
;; ------------------------------------------------------------

(defmacro col [colname]
  (import [cheat-importing.common [amcc-fields]] [hy [HyInteger]])
  `(get row ~(HyInteger (.index amcc-fields (string colname)))))

(defn nie [x]
  ; None if empty
  (if (or (= (len x) 0) (= x "¿¿¿"))
    None
    x))

;; ------------------------------------------------------------
;; * Setup
;; ------------------------------------------------------------

(setv db (sqlite3.connect target-db-path))
(.execute db "pragma foreign_keys = on")
(.execute db "attach ? as Z" [zipstates-db-path])
(unless (.fetchall (.execute db "select name from sqlite_master where type='table' and name='T'"))
  (.execute db "pragma journal_mode = wal")
  (.execute db "create table T (
    cn                 integer primary key,  -- Confirmation number
    usernum            integer not null,
    amount_cents       integer not null,
    time               integer not null,  -- As a POSIX timestamp.
    first_name         text not null,
    last_name          text not null,
    addrnum1           integer not null,
    addrnum2           integer,
    zip                integer not null)"))

;; ------------------------------------------------------------
;; * Main loop for data processing
;; ------------------------------------------------------------

(setv to-insert [])
(setv legal-zipcodes (frozenset (map first (.execute db
  "select zipcode from Z.T"))))
(os.chdir csv-dir)
(setv files (sorted (os.listdir ".")))
(for [[file-i filename] (filt
    (or (not start-with) (>= (second it) start-with))
    (enumerate files))]

  (with [o (open filename :encoding "UTF-8")]
    (print (.format "Reading {:4d}/{:4d} - {}"
      (inc file-i) (len files) filename))
    (setv dialect (.sniff (csv.Sniffer) (.read o (* 5 1024))))
    (.seek o 0)
    (setv reader (csv.reader o dialect))
    ; Skip the first three lines (making sure they're what we think
    ; they are).
    (assert (= (next reader) ["\ufeff"]))
    (assert (= (next reader) []))
    (assert (= (next reader) ["ACCOUNT" "ACCOUNT NAME" "AMOUNT" "AUTH CODE" "AVS" "BRAND" "CARD ENDING" "CVD" "FIRST NAME" "LAST NAME" "MERCHANT TRANS. ID" "OPTION CODE" "DATE" "TXN ID" "CONF. NO." "ERROR CODE" "AUTH TYPE" "TYPE" "TXT_CITY%2CTXT_COUNTRY%2CTXT_EMAIL%2CTXT_PHONE%2CTXT_STATE%2CTXT_ADDR1%2CTXT_ADDR2%2CZIP%2CCONSUMER_IP" ""]))

    (for [row reader]
      (unless (<= (len amcc-fields) (len row) (inc (len amcc-fields)))
        (print "Cut-off line:" (list (zip amcc-fields row)))
        (continue))
      (unless (= (col TYPE) "Settlements")
        ; Only this transaction type should indicate a successful,
        ; cleared transaction.
        (continue))
      (unless (.isdigit (col USERNUM))
        ; If the user number is missing or messed up, I think
        ; this probably wasn't a real transaction.
        (continue))
      (unless (in (col AVS) "DFMXY")
        ; Only these AVS codes indicate a match on street address
        ; and ZIP code.
        ;
        ; I'm assuming the AVS codes are those of PayPal, since
        ; the AM leak has a list of PayPal accounts, and all the
        ; AVS codes in the data (except for 193 instances of Q, and
        ; 892 instances of the null string) are also included in the
        ; first table in https://web.archive.org/web/20150701011128/https://developer.paypal.com/docs/classic/api/AVSResponseCodes/
        (continue))

      (setv zipcode (or (digest-zipcode (col ZIP)) (continue)))
      (unless (in zipcode legal-zipcodes)
        (continue))

      (when (= (col ERROR_CODE) "3203")
        ; The corresponding authorization was either fully
        ; settled or cancelled, so this settlement couldn't go
        ; through.
        (continue))
      (assert (= (col ERROR_CODE) ""))

      (assert (none? (nie (col TXT_ADDR2))))

      (setv name (col NAME))
      (when (re.match r"\A[¿\s]*\Z" name)
        ; There are some strings that are just patterns of ¿ and
        ; spaces. Probably this is the outcome of some encoding
        ; mistakes on AM's part.
        (continue))
      (setv name (digest-personal-name name :one-word False))
      (setv name-parts (.split name " "))
      (unless (>= (len name-parts) 2)
        (continue))
      (setv first-name (get name-parts 0))
      (setv last-name (get name-parts -1))
      (when (or (< (len first-name) 2) (< (len last-name) 2))
        ; Even if a 1-character name is correct in some sense,
        ; it would be tough to unambiguously match.
        (continue))

      (setv amount-cents (do
        (setv x (col AMOUNT))
        (assert (not-in "(" x))
        (assert (not-in "-" x))
        (assert (= (get x -3) "."))
        (int (+
          (.replace (cut x 0 -3) "," "")
          (get x -2)
          (get x -1)))))
      (assert (> amount-cents 0))

      (setv [addrnum1 addrnum2] (get-addrnums (col TXT_ADDR1) zipcode))
      (unless addrnum1
        (continue))

      (.append to-insert (dict
        :cn (int (col CONF_NO))
          ; The confirmation number. TXN_ID might sometimes be a
          ; UUID in hex rather than an integer.
        :usernum (col USERNUM)
        :first_name first-name
        :last_name last-name
        :amount_cents amount-cents
        :time (calendar.timegm (time.strptime (col DATE) "%Y-%m-%d %H:%M:%S"))
        :addrnum1 addrnum1
        :addrnum2 addrnum2
        :zip zipcode))))

  (when (>= (len to-insert) commit-every)
    (commit db to-insert)))

(commit db to-insert)
(.close db)
