(require [kodhy.macros [lc amap whenn block]])

(import
  [os.path :as P]
  [glob [glob]]
  codecs
  re
  csv
  sqlite3
  [unidecode [unidecode]]
  [kodhy.util [T F seq ret]]
  [cheat-importing.common [*]])

(setv voterfiles-dir "../VoterFiles")
(setv zipstates-db-path "../data/zipstates.sqlite")
(setv target-db-path "../data/voter.sqlite")
(setv update-every  10,000)
(setv commit-every 100,000)

; Use these parameters to pick up where a previous run of the
; program left off, or only process a subset of the data.
(setv CA-start-at-line 0)
(setv FL-start-at-file None)
(setv FL-start-at-line 0)
(setv KS-start-at-line 0)
(setv NY-start-at-line 0)
(setv OK-start-at-file "../VoterFiles/OK_2012/VR Info - SW County 28.dat")
(setv OK-start-at-line 0)
(setv done-states ["CA" "FL" "KS" "NY" "OK"])
(setv do-states available-states)

;; ------------------------------------------------------------
;; * Functions and macros
;; ------------------------------------------------------------

(defmacro check [expr]
  `(unless ~expr
    (import [kodhy.util [show-expr]])
    (raise (ValueError (+ "Not true: " (show-expr '~expr))))))

(defn win->ascii [s]
  (unidecode (codecs.decode s "Windows-1252")))

(defn mdy [s]
  ; Converts a string like "02/15/1989" to an integer like 19891502.
  (when s
    (assert (= (len s) (len "02/15/1989")))
    (int (+ (cut s -4) (cut s 0 2) (cut s 3 5)))))

(defn maybe-int [s]
  (when s (int s)))

;; ------------------------------------------------------------
;; * Setup
;; ------------------------------------------------------------

(setv db (sqlite3.connect target-db-path))
(.execute db "attach ? as Z" [zipstates-db-path])
(unless (.fetchall (.execute db "select name from sqlite_master where type='table' and name='T'"))
  (.execute db "pragma journal_mode = wal")
  (.execute db "pragma page_size = 4096")
  (.execute db "create table T (
    rn integer primary key,
      -- The state-specific voter ID number plus an integer code for
      -- the state (0 for AL, 1 for AK, etc.) which itself is
      -- multiplied by 1e12 (so rn values for AK are from
      -- 1,000,000,000,000 to 1,999,999,999,999).
    first_name text not null,
    last_name text not null,
    addrnum1 integer not null,
    addrnum2 integer,
    zip integer not null,
    party integer not null,
    born integer,
    registered integer,
      -- 'born' and 'registered' are dates in the form YYYYmmdd.
    female integer,
      -- 0 (male) or 1 (female).
    gender_is_imputed integer not null default 0,
      -- 0 (not imputed) or 1 (imputed).
    am_user integer)
      -- Initially null; filled in by
      -- cheat-importing.match-am-to-voter"))
(setv max-addrnums 2)

;; ------------------------------------------------------------
;; * Main loop for data processing
;; ------------------------------------------------------------

(setv to-insert [])
(setv current-file None)
(setv line-n 0)
(setv row None)
(try

  (for [state available-states]

    (unless (and (in state do-states) (not-in state done-states))
      (continue))
    (for [st available-states]
      (setv (get (globals) st) (= state st)))
    (setv statenum (state-code state))
    (setv legal-zipcodes (frozenset (map first (.execute db
      "select zipcode from Z.T where state = ?" [(.index states state)]))))
    (check legal-zipcodes)
    (setv exact-n-fields (when KS
      (+ (len (get voter-fields "KS")) 1)))
      ; For Kansas, there are only a handful of rows without this
      ; exact number of fields, which appears to be due to column
      ; misalignment.
    (setv min-n-fields (cond
      [CA (inc (.index (get voter-fields "CA") "Previous Name Prefix"))]
      [KS None]
      [OK (inc (.index (get voter-fields "OK") "Mailing address ZIP Code"))]
      [T (len (get voter-fields state))]))
    (setv CA-min-len-for-nsa
      (inc (.index (get voter-fields "CA") "Non Standard Address")))

    (defmacro field [state inp]
      (defn getrow [x]
        (import [cheat-importing.common [voter-fields]] [kodhy.util [keyword->str]])
        `(get row ~(.index (get voter-fields (keyword->str state)) x)))
      (if (string? inp)
        (getrow inp)
        `(, ~@(amap (getrow it) inp))))

    (defmacro stcase [&rest args]
      (import [kodhy.util [pairs]])
      `(cond
        ~@(lc [[k expr] (apply pairs args)]
          `[~(HySymbol k) ~expr])
        [T (raise (ValueError (+ "unknown state " state)))]))

    (setv files (sorted (do
      (setv d (P.join voterfiles-dir (+ state "_2012")))
      (stcase
        :CA [(P.join d "psr00978.txt")]
        :FL (glob (P.join d "VoterExtract" "*"))
        :KS [(P.join d "Statewide.txt")]
        :NY [(P.join d "NY_voter_file_2012.csv")]
        :OK (glob (P.join d "VR *.dat"))))))
    (setv start-at-file (cond
      [FL FL-start-at-file]
      [OK OK-start-at-file]))

    (for [current-file files]
      (when (and start-at-file (< current-file start-at-file))
        (continue))
      (print current-file)
      (with [o (open current-file :encoding "Windows-1252")]
        (setv reader (csv.reader o
          :quoting (if OK csv.QUOTE-NONE csv.QUOTE-MINIMAL)
          :delimiter (str (cond
            [NY ","]
            [OK ";"]
            [T "\t"]))))
        (setv line-n 0)
        (when (or CA KS)
          (check (= (amap (.strip it) (next reader)) (get voter-fields state)))
          (+= line-n 1))
        (when NY
          (check (= (next reader) (amap (.format "V{}" it) (seq 1 45))))
          (+= line-n 1))
        (setv start-at-line (stcase
          :CA CA-start-at-line
          :FL (if (= current-file start-at-file)
            FL-start-at-line
            0)
          :KS KS-start-at-line
          :NY NY-start-at-line
          :OK (if (= current-file start-at-file)
            OK-start-at-line
            0)))
        (setv iterator (if NY
          ; There's a mis-encoded line we need to catch.
          ((fn []
            (while True
              (try
                (yield (next reader))
                (except [UnicodeDecodeError]
                  (print "Skipping row due to encoding error"))))))
          ; Otherwise, we can use the reader directly.
          reader))
        (for [row iterator]

          (+= line-n 1)
          (when (= (% line-n update-every) 0)
            (print (.format "{} {:,}" current-file line-n)))
          (when (< line-n start-at-line)
            (continue))
          (when (and
              (= (% line-n commit-every) 0)
              (!= line-n start-at-line))
            (commit db to-insert))
          (unless (if exact-n-fields
              (= (len row) exact-n-fields)
              (>= (len row) min-n-fields))
            (continue))
          (when (and FL
              (= (field :FL "Suppress Address, DOB and Phone Number") "Y"))
            (continue))

;; ** Registration number

         (when (and NY (= (field :NY "sboeid") ""))
            (continue))
         (when (and OK (= (field :OK "Voter identification number") "E"))
           ; There are 2 spurious cells here, which result in
           ; column misalignment. Remove them.
           (for [_ (range 2)]
             (.pop row (.index (get voter-fields state) "Voter identification number"))))
         (setv rn (int (stcase
           :CA (field :CA "Registrant ID")
           :FL (field :FL "Voter ID")
           :NY (.lstrip (field :NY "sboeid") "NY")
           :KS (field :KS "text_registrant_id")
           :OK (field :OK "Voter identification number"))))
         (assert (< rn 1,000,000,000))
         (when CA
           ; We need to include the locality code because registrant
           ; IDs are reused across localities. We'll encode it as
           ; billions.
           (setv lc (int (field :CA "Locality Code")))
           (assert (< lc 1,000))
           (+= rn (* 1,000,000,000 lc)))
         ; Now add the state number (in the trillions).
         (+= rn statenum)

;; ** First and last names

          (setv first-name (digest-personal-name :one-word T (stcase
            :CA (field :CA "First Name")
            :FL (field :FL "Name First")
            :KS (field :KS "text_name_first")
            :NY (field :NY "firstname")
            :OK (field :OK "First name"))))
          (setv last-name (digest-personal-name :one-word T (stcase
            :CA (field :CA "Last Name")
            :FL (field :FL "Name Last")
            :KS (field :KS "text_name_last")
            :NY (field :NY "lastname")
            :OK (field :OK "Last name"))))
          (when (or (< (len first-name) 2) (< (len last-name) 2))
            ; Even if a 1-character name is correct in some sense,
            ; it would be tough to unambiguously match.
            (continue))

;; ** ZIP code

          (setv zipcode None)
          (for [z (stcase
              :CA (field :CA ["Zip" "Mailing Zip" "Previous Residence Zip"])
              :FL (field :FL ["Residence Zipcode" "Mailing Zipcode"])
              :KS (field :KS ["text_res_zip5" "text_mail_zip5"])
              :NY (field :NY ["rzip5"])
              :OK (field :OK ["ZIP Code" "Mailing address ZIP Code"]))]
            (setv z (digest-zipcode z))
            (unless (in z legal-zipcodes)
              (continue))
            (setv zipcode z)
            (break))
          (unless zipcode
            (continue))

;; ** Address numbers

          (setv [addrnum1 addrnum2] (get-addrnums
            :zipcode zipcode
            (.join "   " (stcase
              :CA (+
                (field :CA [
                  "Addr. Num."
                  "Mailing Address1" "Mailing Address2"
                  "Previous Residence Street Number"])
                (if (>= (len row) CA-min-len-for-nsa)
                  (, (field :CA "Non Standard Address"))
                  (,)))
              :FL (field :FL [
                "Residence Address Line 1" "Residence Address Line 2"
                "Mailing Address Line 1" "Mailing Address Line 2"
                "Mailing Address Line 3"])
              :KS (field :KS [
                "text_res_address_nbr"
                "text_mail_address1" "text_mail_address2"
                "text_mail_address3" "text_mail_address4"])
              :NY (field :NY [
                "raddnumber"
                "mailadd1" "mailadd2" "mailadd3" "mailadd4"])
              :OK (field :OK [
                "Street/house number"
                "Mailing street address 1" "Mailing street address 2"])))))
          (unless addrnum1
            (continue))

;; ** Party

          (setv party (stcase
            :CA (.rstrip (field :CA "Party"))
            :FL (.upper (field :FL "Party Affiliation"))
            :KS (field :KS "desc_party")
            :NY (if (= (field :NY "enrollment") "OTH")
              (field :NY "otherparty")
              (field :NY "enrollment"))
            :OK (field :OK "Political affiliation")))
          (when (and OK (= party "IND"))
            ; For Oklahoma, there are only three party codes
            ; (plus a few cases marked AE), so IND presumably
            ; means no party.
            (setv party "np"))
;          (unless (in party party-codes)
;            (setv (get party-codes party) (inc (max (.values party-codes))))
;            (print "••• NEW PARTY: " party (get party-codes party)))

;; ** Dates of birth and registration

          (setv born (stcase
            :CA (mdy (field :CA "Date of Birth"))
            :FL (mdy (field :FL "Birth Date"))
            :KS (mdy (field :KS "date_of_birth"))
            :NY (maybe-int (field :NY "dob"))
            :OK (maybe-int (field :OK "Birth date"))))
          (setv registered (stcase
            :CA (mdy (field :CA "Registration Date"))
            :FL (mdy (field :FL "Registration Date"))
            :KS (mdy (field :KS "date_of_registration"))
            :NY (maybe-int (field :NY "regdate"))
            :OK (maybe-int (field :OK "Date of registration"))))

;; ** Gender

          (setv female None)
          (unless OK ; OK does not provide gender.
            (setv gender (.upper (.rstrip (stcase
              :CA (field :CA "Gender")
              :FL (field :FL "Gender")
              :KS (field :KS "cde_gender")
              :NY (field :NY "gender")))))
            (setv female (cond
              [(= gender "M") 0]
              [(= gender "F") 1]
              [(in gender ["" "U"]) None]
              [T (raise (ValueError (.format "gender {!r}" gender)))])))

;; ** Put it all together.

          (.append to-insert (dict
            :rn rn
            :first_name first-name
            :last_name last-name
            :addrnum1 addrnum1
            :addrnum2 addrnum2
            :zip zipcode
            :party (get party-codes party)
            :born born
            :registered registered
            :female female)))

      (commit db to-insert))))

;; ** Exception handling

  (except [BaseException]
    (print (.format "Exception at file {}, line {}: {}"
      current-file
      line-n
      (if row (list (zip (get voter-fields state) row)) "[not read]")))
    (raise)))

;; ------------------------------------------------------------
;; * Finishing touches
;; ------------------------------------------------------------

(when (= (.union (set do-states) (set done-states)) (set available-states))

  (print "Changing journal mode")
  (.execute db "pragma journal_mode = delete")

;; ** Impute genders.

  (print "Creating gender-imputation table")
  (setv male-p .05)
  (setv female-p .95)
  (.execute db "create table GN as select
     first_name,
     (sum(female) + 1)/(count(*) + 2.) as p_female
     from T
     where female notnull
     group by first_name
     having p_female <= ? or p_female >= ?" (, male-p female-p))
      ; p_female is calculated as the posterior mean of a Bernoulli
      ; distribution with a Beta(1, 1) prior.
  (for [female [0 1]]
    (print (.format "Imputing genders ({})" female))
    (.execute db (.format "
      update T
        set gender_is_imputed = 1, female = {}
        where female isnull and
          first_name in (select first_name from GN where {})"
      female (if female "p_female > .5" "p_female < .5"))))
  (commit db [])

;; ** Create the index.

  (print "Creating index")
  (.execute db "create index Tnamezip on T (first_name, last_name, zip)")
  (commit db []))
