(require [kodhy.macros [rmap amap filt $ geta wc ss]])

(import
  random
  [numpy :as np]
  scipy.special
  [pandas :as pd]
  patsy
  sklearn.linear-model
  [kodhy.util [T F seq concat by-2s drop-unused-cats rd]])

(defn multiset-sample [counts sample-size]
  ; Takes a list or 1D array of counts and a sample size.
  ; Samples from `counts` without replacement.
  ; Returns a 1D array of the same length as `counts`, whose sum
  ; is equal to `sample-size`.
  ; http://stackoverflow.com/a/35735195
  (setv remaining (cut (np.cumsum (cut counts None None -1)) None None -1))
  (setv result (np.zeros (len counts) :dtype np.int))
  (for [i (range (dec (len counts)))]
    (when sample-size
      (setv (get result i) (np.random.hypergeometric
        (get counts i) (get remaining (+ i 1)) sample-size))
      (-= sample_size (get result i))))
  (setv (get result -1) sample-size)
  result)

(defn gen-counts-cv-folds [counts &optional [n-folds 10]]
  (import [numpy :as np])
  ; Determine fold sizes.
  (setv total (np.sum counts))
  (setv fold-sizes (np.repeat (// total n-folds) n-folds))
  (+= (geta fold-sizes (list (range (% total n-folds)))) 1)
    ; This tops up the folds with any leftover cases.
  ; Construct 'folds' as a list of arrays of counts.
  (setv remaining (np.array counts))
  (setv folds [])
  (for [n (cut fold-sizes None -1)]
    (.append folds (multiset-sample remaining n))
    (-= remaining (get folds -1)))
  (.append folds remaining)
  folds)

(defn gen-counts-cv-folds-ts [trials successes &optional [n-folds 10]]
  ; A wrapper for `gen-counts-cv-folds` that handles
  ; trials-and-successes format, converting it to a
  ; failures-and-successes format and back.
  ;
  ; `gen-counts-cv-folds` doesn't actually need to distigunish
  ; between failures and successes, so we concatenate them all
  ; into one big array.
  (setv counts (np.array (concat (rmap [[t s] (zip trials successes)]
    [(- t s) s]))))
  (setv folds (gen-counts-cv-folds counts n-folds))
  (rmap [fold folds]
    (setv [fs ss] (amap (np.array it) (zip #* (by-2s fold))))
    (, (+ fs ss) ss)))

(defn trials->failures [trials successes]
  (np.array (lc [[t s] (zip trials successes)] (- t s))))

(defn failures-x-y-counts [x-mat trials successes]
  (setv failures (trials->failures trials successes))
  (setv x (np.vstack (, x-mat x-mat)))
  (setv y (np.append
    (np.repeat 0 (get x-mat.shape 0))
    (np.repeat 1 (get x-mat.shape 0))))
  (setv counts (np.append failures successes))
  [x y counts])

(defn counts-brier [prob-pred trials successes]
  ; First, filter out rows for which there are no trials.
  ; Those rows provide no ground truth with which to test the
  ; prediction.
  (setv prob-pred (get prob-pred (!= trials 0)))
  (setv successes (get successes (!= trials 0)))
  (setv trials (get trials (!= trials 0)))
  ; Now compute SSE and MSE.
  (setv sse (np.sum (+
    (* (** (- prob-pred 1) 2) successes)
    (* (**    prob-pred    2) (- trials successes)))))
  (/ sse (np.sum trials)))

(defn counts-cond-prob [prob-pred trials successes]
  (setv failures (- trials successes))
  (,
    (/ (np.sum (* prob-pred failures)) (np.sum failures))
    (/ (np.sum (* prob-pred successes)) (np.sum successes))))

(defn counts-mean-sd [x trials]
  (setv m (/ (np.sum (* x trials)) (np.sum trials)))
  (setv sd (np.sqrt (/ (np.sum (* (** (- x m) 2) trials)) (np.sum trials))))
  [m sd])

(defn counts-cv-pred [x trials successes f &optional folds [n-folds 10] [verbose None]]
"f will generally be of the form
(fn [x train-trials train-successes] ...)
and should return a 1D nparray of a prediction for each row of x.

Predictions are aggregated across folds with means weighted by
the count of the test trials."
  (import [numpy :as np])
  (assert (= (len trials) (len successes)))
  (unless folds
    (setv folds (gen-counts-cv-folds-ts trials successes n-folds)))
  (setv fold-preds (np.column-stack (rmap [[i [test-trials test-successes]] (enumerate folds)]
    (when verbose
      (print (.format "{}: Fold {}" verbose (+ i 1))))
    (setv train-trials (- trials test-trials))
    (setv train-successes (- successes test-successes))
    (* (f x train-trials train-successes) test-trials))))
  (/ (np.sum fold-preds 1) trials))

(defn patsy-no-intercept [formula data]
  ; Takes a formula with an intercept and returns a design
  ; matrix without one, but also with factors coded as if there
  ; were an intercept. This is useful for input to scikit-learn
  ; functions that treat the intercept specially.
  (setv x (patsy.dmatrix formula data))
  (assert (= (.name (get x.design-info.terms 0)) "Intercept"))
  (setv new-di (.subset x.design-info (cut x.design-info.terms 1)))
  (setv x (np.delete x 0 1))
  (setv x.design-info new-di)
  x)

(defn patsy-drop-constants [x]
  ; Removes constant columns. The result could have a complex
  ; set of Patsy terms, so it's not a real Patsy design matrix,
  ; just a matrix with a list of columns.
  (setv ixs (filt
    (= (len (np.unique (geta x (~ (np.isnan (geta x : it))) it))) 1)
    (range (second x.shape))))
  (setv new-col-names (lc
    [[i t] (enumerate x.design-info.column-names)]
    (not-in i ixs)
    t))
  (setv x (np.delete (np.matrix x) (np.array ixs) 1))
  (setv x.column-names new-col-names)
  x)

(defn counts-logitreg [mode x trials successes C]
  (setv m (sklearn.linear-model.LogisticRegression
    :solver "lbfgs" :C C :max-iter 1e100))
  (setv [x2 y sfcounts] (failures-x-y-counts x trials successes))
  (.fit m x2 y :sample-weight sfcounts)
  (cond
    [(= mode "predict")
      (geta (.predict-proba m x) : 1)]
    [(= mode "coef")
      (pd.Series
        (+ (list m.intercept_) (list (first m.coef_)))
        (+ ["Intercept"] x.column-names))]
    [T
      (raise (ValueError))]))

(setv n-folds-Ccv 5)
(defn counts-logitreg-Ccv [mode x trials successes]

  ; Choose a value of C, the regularization strength, with
  ; cross-validation.
  (setv Cs (** 10 (np.arange -12. 13. 1)))
  (setv C-folds (gen-counts-cv-folds-ts
    trials successes n-folds-Ccv))
  (setv briers (np.array (rmap [C Cs]
    (setv p-pred (counts-cv-pred
      x trials successes :folds C-folds
      (fn [inner-x inner-trials inner-successes]
        (counts-logitreg "predict"
          inner-x inner-trials inner-successes
          :C C))))
    (counts-brier p-pred trials successes))))
  (setv C (get Cs (np.argmin briers)))
  (print "Chose C =" C)

  ; Now fit to all the data with the selected C.
  (counts-logitreg mode x trials successes :C C))

(defn am-usage-filter [df]
  (drop-unused-cats (ss df (&
    (.isin $party keep-parties)
    (.notnull $gender)
    (.notnull $age)))))

(defn am-usage-transform-ages [reference-data input-ages]
  ; Create a squared age term, and standardize age and squared age.
  (setv df reference-data)
  (setv [age-mean age-sd] (counts-mean-sd ($ df age) ($ df voters)))
  (setv age-t (/ (- input-ages age-mean) (* 2 age-sd)))
  (setv [age2-mean age2-sd] (counts-mean-sd (** (- ($ df age) age-mean) 2) ($ df voters)))
  (setv age2-t (/ (- (** (- input-ages age-mean) 2) age2-mean) (* 2 age2-sd)))
  [age-t age2-t])

(defn am-usage [mode df formula &optional folds [trivial-model F]]

  (setv df (am-usage-filter df))
  (setv ($ df female) (wc df (& (.notnull $gender) (= $gender "Female"))))
  ; We make New York the first category in the 'state' column so
  ; that it becomes the reference category.
  (.reorder-categories (. ($ df state) cat)
    (+ ["NY"] (filt (!= it "NY") (. ($ df state) cat categories)))
    :inplace T)
  (setv [($ df age) ($ df age2)] (am-usage-transform-ages df ($ df age)))

  (setv xm (patsy-drop-constants (patsy-no-intercept formula df)))

  (setv trials (.as-matrix ($ df voters)))
  (setv successes (.as-matrix ($ df am_users)))

  (cond

    [(= mode "fit_all") (do
      (assert (not trivial-model))
      (counts-logitreg-Ccv "coef" xm trials successes))]

    [(= mode "cv") (do
      (assert folds)
      (setv p-pred (counts-cv-pred xm trials successes :verbose "outer" :folds folds
        (fn [x train-trials train-successes]
          (if trivial-model
            (np.repeat (/ (np.sum train-successes) (np.sum train-trials)) (get x.shape 0))
            (counts-logitreg-Ccv "predict" x train-trials train-successes)))))
      {
        :x-column-names xm.column-names
        :cv-mse (counts-brier p-pred trials successes)
        :cv-p-cond (counts-cond-prob p-pred trials successes)})]))
